import argparse
import logging
import os

import h5py
import numpy as np
import pandas as pd
from scipy.io import loadmat
from temporaldata import (
    ArrayDict,
    Data,
    Interval,
    IrregularTimeSeries,
)

from brainsets import serialize_fn_map
from brainsets.descriptions import (
    BrainsetDescription,
    SessionDescription,
    SubjectDescription,
    DeviceDescription,
)
from brainsets.taxonomy import (
    RecordingTech,
    Species,
    Task,
    Sex,
)


logging.basicConfig(level=logging.INFO)


def extract_units(mat):
    """
    Get unit metadata for the session.
    This is the same for all trials.

    ..note::
        Currently only populating unit_name and unit_number.
    """
    values = mat["Subject"][0][0][0]

    # Only get unit meta from the first trial
    unit_meta = []
    neurons = values["Neuron"][0][0]
    for i, _ in enumerate(neurons):
        unit_name = f"unit_{i}"
        unit_meta.append(
            {
                "id": unit_name,
                "unit_number": i,
            }
        )

    units = ArrayDict.from_dataframe(pd.DataFrame(unit_meta))
    return units


def extract_behavior(mat, trials):
    """
    Get lists of behavior timestamps and hand velocity for each trial.
    These timestamps are regularly sampled every 100ms.

    Parameters:
    - mat: A dictionary containing the data extracted from a MATLAB file.

    Returns:
    - behavior object of type IrregularTimeSeries
    """
    values = mat["Subject"][0][0][0]
    behavior_timestamps_list = []
    hand_vel_list = []
    for trial_id in range(len(values["Time"])):
        behavior_timestamps = values["Time"][trial_id][0][:, 0]
        hand_vel = values["HandVel"][trial_id][0][:, :2]
        behavior_timestamps_list.append(behavior_timestamps)
        hand_vel_list.append(hand_vel)

    behavior_timestamps = np.concatenate(behavior_timestamps_list).astype(np.float64)
    hand_vel = np.concatenate(hand_vel_list)

    hand = IrregularTimeSeries(
        timestamps=behavior_timestamps,
        vel=hand_vel,
        domain=trials,
    )
    return hand


def extract_spikes(mat):
    """
    Extracts spike timestamps and unit ids for each trial from a MATLAB file.

    Parameters:
    - mat: A dictionary containing the data extracted from a MATLAB file.

    Returns:
    - spikes object of type IrregularTimeSeries


    """
    values = mat["Subject"][0][0][0]
    spike_timestamps_list = []
    unit_id_list = []
    trial_start = []
    trial_end = []
    for trial_id in range(len(values["Time"])):
        neurons = values["Neuron"][trial_id][0]
        tstart = np.inf
        tend = 0
        for i, neuron in enumerate(neurons):
            spiketimes = neuron[0][0]
            if len(spiketimes) == 0:
                continue
            spiketimes = spiketimes[:, 0]
            spike_timestamps_list.append(spiketimes)
            unit_id_list.append(np.ones_like(spiketimes, dtype=np.int64) * i)
            tstart = spiketimes.min() if spiketimes.min() < tstart else tstart
            tend = spiketimes.max() if spiketimes.max() > tend else tend
        trial_start.append(tstart)
        trial_end.append(tend)

    spikes = np.concatenate(spike_timestamps_list).astype(np.float64)
    unit_ids = np.concatenate(unit_id_list)
    spikes = IrregularTimeSeries(
        timestamps=spikes,
        unit_index=unit_ids,
        domain=Interval(
            np.array(trial_start, dtype=np.float64),
            np.array(trial_end, dtype=np.float64),
        ),
    )
    spikes.sort()
    assert spikes.domain.is_disjoint()
    assert spikes.domain.is_sorted()
    return spikes


def extract_trials(mat):
    """
    Get a list of trial intervals for each trial.
    These intervals are defined by the start and end of the behavior timestamps.

    Parameters:
    - mat: A dictionary containing the data extracted from a MATLAB file.

    Returns:
    - A list of trial intervals for each trial of type Interval.
    """
    values = mat["Subject"][0][0][0]
    trial_starts = []
    trial_ends = []
    for trial_id in range(len(values["Time"])):
        behavior_timestamps = values["Time"][trial_id][0][:, 0]
        trial_starts.append(behavior_timestamps.min())
        trial_ends.append(behavior_timestamps.max())
    trials = Interval(
        start=np.array(trial_starts, dtype=np.float64),
        end=np.array(trial_ends, dtype=np.float64),
    )
    assert trials.is_disjoint()
    assert trials.is_sorted()
    return trials


def main():
    # Use argparse to extract two arguments from the command line:
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_dir", type=str, default="./processed")

    args = parser.parse_args()

    brainset_description = BrainsetDescription(
        id="flint_slutzky_accurate_2012",
        origin_version="0.0.0",
        derived_version="1.0.0",
        source="https://portal.nersc.gov/project/crcns/download/dream/data_sets/Flint_2012",
        description="Monkeys recordings of Motor Cortex (M1) and dorsal Premotor Cortex"
        " (PMd)  128-channel acquisition system (Cerebus,Blackrock, Inc.)  "
        "while performing reaching tasks on right hand",
    )

    logging.info(f"Processing file: {args.input_file}")

    # open file
    mat = loadmat(args.input_file)

    subject = SubjectDescription(
        id="monkey_c",
        species=Species.MACACA_MULATTA,
        sex=Sex.UNKNOWN,
    )

    session_tag = args.input_file.split("_")[-1].split(".mat")[0]  # e1, e2, e3...
    device_id = f"{subject.id}_{session_tag}"
    session_id = f"{device_id}_reaching"

    # register session
    session = SessionDescription(
        id=session_id,
        recording_date="20130530",  # using .mat file creation date
        task=Task.REACHING,
    )

    device = DeviceDescription(
        id=device_id,
        recording_tech=RecordingTech.UTAH_ARRAY_SPIKES,
    )

    units = extract_units(mat)  # Data obj

    spikes = extract_spikes(mat)  # IrregularTimeSeries obj

    trials = extract_trials(mat)  # Interval obj

    hand = extract_behavior(mat, trials)  # IrregularTimeSeries obj

    data = Data(
        brainset=brainset_description,
        subject=subject,
        session=session,
        device=device,
        # neural activity
        spikes=spikes,
        units=units,
        # stimuli and behavior
        trials=trials,
        hand=hand,
        domain=trials,
    )

    # split trials into train, validation and test
    train_trials, valid_trials, test_trials = trials.split(
        [0.7, 0.1, 0.2], shuffle=True, random_seed=42
    )

    data.set_train_domain(train_trials)
    data.set_valid_domain(valid_trials)
    data.set_test_domain(test_trials)

    # save data to disk
    path = os.path.join(args.output_dir, f"{session_id}.h5")
    with h5py.File(path, "w") as file:
        data.to_hdf5(file, serialize_fn_map=serialize_fn_map)


if __name__ == "__main__":
    main()



import click

from .utils import get_available_brainsets


@click.command()
def cli_list():
    """List available brainsets."""
    click.echo("Available brainsets:")
    for brainset in get_available_brainsets():
        click.echo(f"- {brainset}")



import numpy as np
import pandas as pd

from temporaldata import ArrayDict, IrregularTimeSeries

from brainsets.descriptions import SubjectDescription
from brainsets.taxonomy import (
    RecordingTech,
    Sex,
    Species,
)


def extract_metadata_from_nwb(nwbfile):
    recording_date = nwbfile.session_start_time.strftime("%Y%m%d")
    related_publications = nwbfile.related_publications
    return dict(
        recording_date=recording_date, related_publications=related_publications
    )


def extract_subject_from_nwb(nwbfile):
    r"""DANDI has requirements for metadata included in `subject`. This includes:
    subject_id: A subject identifier must be provided.
    species: either a latin binomial or NCBI taxonomic identifier.
    sex: must be "M", "F", "O" (other), or "U" (unknown).
    date_of_birth or age: this does not appear to be enforced, so will be skipped.
    """
    species = nwbfile.subject.species

    if "NCBITaxon" in species:
        species = "NCBITaxon_" + species.split("_")[-1]

    return SubjectDescription(
        id=nwbfile.subject.subject_id.lower(),
        species=Species.from_string(species),
        sex=Sex.from_string(nwbfile.subject.sex),
    )


def extract_spikes_from_nwbfile(nwbfile, recording_tech):
    # spikes
    timestamps = []
    unit_index = []

    # units
    unit_meta = []

    units = nwbfile.units.spike_times_index[:]
    electrodes = nwbfile.units.electrodes.table

    # all these units are obtained using threshold crossings
    for i in range(len(units)):
        if recording_tech == RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS:
            # label unit
            group_name = electrodes["group_name"][i]
            unit_id = f"group_{group_name}/elec{i}/multiunit_{0}"
        elif recording_tech == RecordingTech.UTAH_ARRAY_SPIKES:
            # label unit
            electrode_id = nwbfile.units[i].electrodes.item().item()
            group_name = electrodes["group_name"][electrode_id]
            unit_id = f"group_{group_name}/elec{electrode_id}/unit_{i}"
        else:
            raise ValueError(f"Recording tech {recording_tech} not supported")

        # extract spikes
        spiketimes = units[i]
        timestamps.append(spiketimes)

        if len(spiketimes) > 0:
            unit_index.append([i] * len(spiketimes))

        # extract unit metadata
        unit_meta.append(
            {
                "id": unit_id,
                "unit_number": i,
                "count": len(spiketimes),
                "type": int(recording_tech),
            }
        )

    # convert unit metadata to a Data object
    unit_meta_df = pd.DataFrame(unit_meta)  # list of dicts to dataframe
    units = ArrayDict.from_dataframe(
        unit_meta_df,
        unsigned_to_long=True,
    )

    # concatenate spikes
    timestamps = np.concatenate(timestamps)
    unit_index = np.concatenate(unit_index)

    # create spikes object
    spikes = IrregularTimeSeries(
        timestamps=timestamps,
        unit_index=unit_index,
        domain="auto",
    )

    # make sure to sort ethe spikes
    spikes.sort()

    return spikes, units



"""Signal processing functions. Inspired by Stavisky et al. (2015).

https://dx.doi.org/10.1088/1741-2560/12/3/036009
"""

from typing import List, Tuple

import numpy as np
import tqdm
from scipy import signal

from temporaldata import Data, IrregularTimeSeries, ArrayDict
from brainsets.taxonomy import RecordingTech


def downsample_wideband(
    wideband: np.ndarray,
    timestamps: np.ndarray,
    wideband_Fs: float,
    lfp_Fs: float = 1000,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Downsample wideband signal to LFP sampling rate.
    """
    assert wideband.shape[0] == timestamps.shape[0], "Time should be first dimension."
    # Decimate by a factor of 4
    dec_factor = 4
    if wideband.shape[0] % dec_factor != 0:
        wideband = wideband[: -(wideband.shape[0] % dec_factor), :]
        timestamps = timestamps[: -(timestamps.shape[0] % dec_factor)]
    wideband = wideband.reshape(-1, dec_factor, wideband.shape[1])
    wideband = wideband.mean(axis=1)

    timestamps = timestamps[::dec_factor]

    nyq = 0.5 * wideband_Fs / dec_factor  # Nyquist frequency
    cutoff = 0.333 * lfp_Fs  # remove everything above 170 Hz.
    normal_cutoff = cutoff / nyq
    b, a = signal.butter(4, normal_cutoff, btype="low", analog=False, output="ba")

    # Interpolation to achieve the desired sampling rate
    t_new = np.arange(timestamps[0], timestamps[-1], 1 / lfp_Fs)
    lfp = np.zeros((len(t_new), wideband.shape[1]))
    for i in range(wideband.shape[1]):
        # We do this one channel at a time to save memory.
        broadband_low = signal.filtfilt(b, a, wideband[:, i], axis=0)
        lfp[:, i] = np.interp(t_new, timestamps, broadband_low)

    return lfp, t_new


def extract_bands(
    lfps: np.ndarray, ts: np.ndarray, Fs: float = 1000, notch: float = 60
) -> Tuple[np.ndarray, np.ndarray, List]:
    """Extract bands from LFP

    We prefer to extract bands from the LFP upstream rather than downstream, because
    it can be difficult to estimate e.g. the phase of low-frequency LFPs from
    short segments.

    We use the proposed bands from Stravisky et al. (2015), but we use the MNE toolbox
    rather than straight scipy signal.
    """
    try:
        import mne
    except ImportError:
        raise ImportError(
            "This function requires the MNE library which you can install with "
            "`pip install mne`"
        )

    target_Fs = 50
    assert (
        Fs % target_Fs == 0
    ), "Sampling rate must be a multiple of the target frequency"

    assert lfps.shape[0] == ts.shape[0], "Time should be first dimension."
    info = mne.create_info(
        ch_names=lfps.shape[1], sfreq=Fs, ch_types=["eeg"] * lfps.shape[1]
    )
    data = mne.io.RawArray(lfps.T, info)
    data = data.notch_filter(np.arange(notch, notch * 5 + 1, notch), n_jobs=4)

    filtered = []
    band_names = ["delta", "theta", "alpha", "beta", "gamma", "lmp"]
    bands = [(1, 4), (3, 10), (12, 23), (27, 38), (50, 300)]
    for band_low, band_hi in bands:
        band = data.copy().filter(band_low, band_hi, fir_design="firwin", n_jobs=4)
        band = band.apply_function(lambda x: x**2, n_jobs=4)

        band = band.filter(18, None, fir_design="firwin", n_jobs=4)
        # It seems resample overwrites the original data, so we copy it first.
        band = band.resample(target_Fs, npad="auto", n_jobs=4)

        filtered.append(band.get_data().T)

    lmp = data.copy().filter(0.1, 20, fir_design="firwin", n_jobs=4)
    lmp = lmp.resample(target_Fs, npad="auto", n_jobs=4)
    filtered.append(lmp.get_data().T)

    ts = ts[int(Fs / target_Fs / 2) :: int(Fs / target_Fs)]
    stacked = np.stack(filtered, axis=2)

    # There can be off by one errors.
    if stacked.shape[0] != len(ts):
        stacked = stacked[: len(ts), :, :]

    return stacked, ts, band_names


def cube_to_long(
    ts: np.ndarray, cube: np.ndarray, channel_prefix="chan"
) -> Tuple[List[IrregularTimeSeries], Data]:
    """Convert a cube of threshold crossings to a list of trials and units."""
    assert cube.shape[1] == len(ts)
    assert cube.ndim == 3
    channels = np.arange(cube.shape[2])
    channels = np.tile(channels, [cube.shape[1], 1])

    # First dim is batch, second is time, third is channel.
    assert np.issubdtype(cube.dtype, np.integer)
    assert cube.min() >= 0

    ts = np.tile(ts.reshape((-1, 1)), [1, cube.shape[2]])
    assert ts.shape == channels.shape

    # The first dimension we map to a single trial.
    trials = []
    for b in tqdm.tqdm(range(cube.shape[0])):
        cube_ = cube[b, :, :]
        ts_ = []
        channels_ = []

        # This data is binned, so we create N identifical timestamps when there are N
        # spikes in a bin.
        for n in range(1, cube_.max() + 1):
            ts_.append(ts[cube_ >= n])
            channels_.append(channels[cube_ >= n])

        ts_ = np.concatenate(ts_)
        channels_ = np.concatenate(channels_)

        tidx = np.argsort(ts_)
        ts_ = ts_[tidx]
        channels_ = channels_[tidx]

        trials.append(
            IrregularTimeSeries(
                timestamps=ts_,
                unit_index=channels_,
                types=np.ones(len(ts_))
                * int(RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS),
                domain="auto",
            )
        )

    counts = cube.sum(axis=0).sum(axis=0)
    units = ArrayDict(
        count=np.array(counts.astype(int)),
        channel_name=np.array(
            [f"{channel_prefix}{c:03}" for c in range(cube.shape[2])]
        ),
        unit_number=np.zeros(cube.shape[2]),
        id=np.array([f"{channel_prefix}{c}" for c in range(cube.shape[2])]),
        channel_number=np.arange(cube.shape[2]),
        type=np.ones(cube.shape[2]) * int(RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS),
    )

    return trials, units



from brainsets.core import StringIntEnum


class Orientation_8_Classes(StringIntEnum):
    angle_0 = 0
    angle_45 = 1
    angle_90 = 2
    angle_135 = 3
    angle_180 = 4
    angle_225 = 5
    angle_270 = 6
    angle_315 = 7



from .subject import (
    Species,
    Sex,
)

from .task import (
    Task,
)

from .drifting_gratings import Orientation_8_Classes
from .macaque import Macaque
from .mice import Cre_line

from .recording_tech import (
    RecordingTech,
    Hemisphere,
)



import os
from typing import Callable, List, Optional, Union
import click
import yaml
from pathlib import Path
import brainsets_pipelines
from prompt_toolkit.auto_suggest import AutoSuggest, Suggestion

CONFIG_FILE = Path.home() / ".brainsets.yaml"
PIPELINES_PATH = Path(brainsets_pipelines.__path__[0])


def expand_path(path: Union[str, Path]) -> Path:
    """
    Convert string path to absolute Path, expanding environment variables and user.
    """
    return Path(os.path.abspath(os.path.expandvars(os.path.expanduser(path))))


def load_config(path: Path = CONFIG_FILE, raise_cli_error: bool = True):
    if path.exists():
        with open(path, "r") as f:
            ret = yaml.safe_load(f)

        if raise_cli_error:
            _validate_config(ret)
        else:
            try:
                _validate_config(ret)
            except:
                return None

        return ret
    elif raise_cli_error:
        raise click.ClickException(
            f"Config not found at {path}. Please run `brainsets config`"
        )
    else:
        return None


def _validate_config(config: dict):
    if "raw_dir" not in config:
        raise click.ClickException(
            "'raw_dir' missing in config. Please run `brainsets config`."
        )
    if "processed_dir" not in config:
        raise click.ClickException(
            "'processed_dir' missing in config. Please run `brainsets config`."
        )


def save_config(config):
    with open(CONFIG_FILE, "w") as f:
        yaml.safe_dump(config, f, default_flow_style=False)
    return CONFIG_FILE


def get_available_brainsets():
    ret = [d.name for d in PIPELINES_PATH.iterdir() if d.is_dir()]
    ret = [name for name in ret if not name.startswith((".", "_"))]
    return ret


def debug_echo(msg: str, enable: bool):
    if enable:
        click.echo(f"DEBUG: {msg}")



from enum import Enum
import datetime


class NestedEnumType(type(Enum)):
    def __new__(cls, clsname, bases, clsdict, parent=None):
        new_cls = super().__new__(cls, clsname, bases, clsdict)
        new_cls._parent = parent

        if parent is not None:
            parent._parent_cls = new_cls
            for name, member in new_cls.__members__.items():
                parent.__setattr__(name, member)

        return new_cls

    def __contains__(cls, member):
        return (isinstance(member, cls) and (member._name_ in cls._member_map_)) or (
            member._parent is not None and member._parent in cls
        )


class StringIntEnum(Enum, metaclass=NestedEnumType):
    r"""Base class for string-integer enums.

    This class extends Python's built-in Enum class to provide:
        - String representation via __str__
        - Integer representation via __int__
        - Case-insensitive string parsing via from_string()
        - Maximum value lookup via max_value()

    .. code-block:: python

        >>> class Color(StringIntEnum):
        ...     RED = 1
        ...     BLUE = 2
        >>> str(Color.RED)
        'RED'
        >>> int(Color.RED)
        1
        >>> Color.from_string("red")
        <Color.RED: 1>
        >>> Color.max_value()
        2
    """

    def __str__(self):
        if self._parent is not None:
            return f"{str(self._parent)}.{self.name}"
        else:
            return self.name

    def __int__(self):
        return self.value

    @classmethod
    def from_string(cls, string: str) -> "StringIntEnum":
        r"""Convert a string to an enum member. This method is case insensitive and
        will replace spaces with underscores.

        Args:
            string: The string to convert to an enum member.

        Examples:
            >>> from brainsets.taxonomy import Sex
            >>> Sex.from_string("Male")
            <Sex.MALE: 1>
            >>> Sex.from_string("M")
            <Sex.MALE: 1>
        """
        nested_string = string.split(".", maxsplit=1)
        if len(nested_string) > 1:
            parent = cls.from_string(nested_string[0])
            return parent._parent_cls.from_string(nested_string[1])
        else:
            # normalize string by replacing spaces with underscores and converting
            # to upper case
            normalized_string = string.strip().upper().replace(" ", "_")
            # create a mapping of enum names to enum members
            mapping = {name.upper(): member for name, member in cls.__members__.items()}
            # try to match the string to an enum name
            if normalized_string in mapping:
                return mapping[normalized_string]
            # if there is no match raise an error
            raise ValueError(
                f"{normalized_string} does not exist in {cls.__name__}, "
                "consider adding it to the enum."
            )

    @classmethod
    def max_value(cls):
        r"""Return the maximum value in the enum class."""
        return max(cls.__members__.values(), key=lambda x: x.value).value


class Dictable:
    r"""A dataclass that can be converted to a dict."""

    def to_dict(self):
        r"""Convert the dataclass instance to a dictionary.

        Returns:
            dict: A dictionary containing all fields of the dataclass as key-value pairs.

        .. code-block:: python

            >>> from dataclasses import dataclass
            >>> @dataclass
            ... class Person(Dictable):
            ...     name: str
            ...     age: int

            >>> p = Person("Alice", 30)
            >>> p.to_dict()
            {'name': 'Alice', 'age': 30}
        """
        from dataclasses import asdict

        return {k: v for k, v in asdict(self).items()}  # type: ignore


def string_int_enum_serialize_fn(obj, serialize_fn_map=None):
    r"""Convert a StringIntEnum object to a string."""
    return str(obj)


def datetime_serialize_fn(obj, serialize_fn_map=None):
    r"""Convert a datetime object to a string."""
    return str(obj)


serialize_fn_map = {
    StringIntEnum: string_int_enum_serialize_fn,
    datetime.datetime: datetime_serialize_fn,
}



import argparse
import datetime
import logging
import h5py
import os

import numpy as np
from pynwb import NWBHDF5IO
from scipy.ndimage import binary_dilation, binary_erosion
from sklearn.model_selection import train_test_split

from temporaldata import Data, IrregularTimeSeries, Interval
from brainsets.descriptions import (
    BrainsetDescription,
    SessionDescription,
    DeviceDescription,
)
from brainsets.utils.dandi_utils import (
    extract_spikes_from_nwbfile,
    extract_subject_from_nwb,
)
from brainsets.taxonomy import RecordingTech, Task
from brainsets import serialize_fn_map

logging.basicConfig(level=logging.INFO)


def extract_behavior(nwbfile):
    """Extract behavior from the NWB file.

    ..note::
        Cursor position and target position are in the same frame of reference.
        They are both of size (sequence_len, 2).
    """
    timestamps = nwbfile.processing["behavior"]["Position"]["cursor_pos"].timestamps[:]
    cursor_pos = nwbfile.processing["behavior"]["Position"]["cursor_pos"].data[:]  # 2d
    cursor_vel = nwbfile.processing["behavior"]["Velocity"]["cursor_vel"].data[:]
    cursor_acc = nwbfile.processing["behavior"]["Acceleration"]["cursor_acc"].data[:]

    cursor = IrregularTimeSeries(
        timestamps=timestamps,
        pos=cursor_pos,
        vel=cursor_vel,
        acc=cursor_acc,
        domain="auto",
    )

    return cursor


def extract_center_out_reaching_trials(nwbfile, cursor):
    r"""Extract trial information from the NWB file. Trials that are flagged as
    "to discard" or where the monkey failed are marked as invalid."""
    trial_table = nwbfile.trials.to_dataframe()

    # infer the trial structure from the trial table
    trial_grid = np.append(
        trial_table.target_on_time.iloc[:],
        min(trial_table.stop_time.iloc[-1] + 1.0, cursor.domain.end[-1]),
    )
    default_value = np.append(
        trial_table.start_time.iloc[0], trial_table.stop_time.values + 1.0
    )
    default_value[1:-1] = np.minimum(
        default_value[1:-1], trial_table.stop_time.values[1:]
    )
    nan_mask = np.isnan(trial_grid)
    trial_grid[nan_mask] = default_value[nan_mask]
    trial_grid = trial_grid.astype(np.float64)
    trial_table["end"] = trial_grid[1:]
    trial_table["start"] = trial_grid[:-1]

    trials = Interval.from_dataframe(trial_table)
    assert trials.is_disjoint()

    # find valid trials
    success_mask = trials.result == "R"
    valid_target_mask = ~np.isnan(trials.target_id)
    max_duration_mask = (trials.end - trials.start) < 6.0
    min_duration_mask = (trials.end - trials.start) > 0.5

    trials.is_valid = (
        success_mask & valid_target_mask & max_duration_mask & min_duration_mask
    )

    valid_trials = trials.select_by_mask(trials.is_valid)

    # isolate movement phases
    movement_phases = Data(
        hold_period=Interval(
            start=valid_trials.target_on_time, end=valid_trials.go_cue_time
        ),
        reach_period=Interval(
            start=valid_trials.go_cue_time, end=valid_trials.stop_time
        ),
        return_period=Interval(start=valid_trials.stop_time, end=valid_trials.end),
        invalid=trials.select_by_mask(~trials.is_valid),
        domain="auto",
    )

    # everything outside of the different identified periods will be marked as random
    movement_phases.random_period = cursor.domain.difference(movement_phases.domain)

    return trials, movement_phases


def extract_random_target_reaching_trials(nwbfile, cursor):
    r"""Extract trial information from the NWB file. Trials that are flagged as
    "to discard" or where the monkey failed are marked as invalid."""
    trial_table = nwbfile.trials.to_dataframe()

    # rename start and end time columns
    trial_table = trial_table.rename(
        columns={
            "start_time": "start",
            "stop_time": "end",
        }
    )

    trials = Interval.from_dataframe(trial_table)

    # find valid trials
    success_mask = trials.result == "R"
    valid_num_attempts = trials.num_attempted == 4
    max_duration_mask = (trials.end - trials.start) < 10.0
    min_duration_mask = (trials.end - trials.start) > 2.0

    trials.is_valid = (
        success_mask & valid_num_attempts & max_duration_mask & min_duration_mask
    )

    valid_trials = trials.select_by_mask(~np.isnan(trials.go_cue_time_array[:, 0]))

    movement_phases = Data(
        hold_period=Interval(
            start=valid_trials.start, end=valid_trials.go_cue_time_array[:, 0]
        ),
        domain="auto",
    )

    # everything outside of the different identified periods will be marked as random
    movement_phases.random_period = cursor.domain.difference(movement_phases.domain)

    return trials, movement_phases


def detect_outliers(cursor):
    # sometimes monkeys get angry, we want to identify the segments where the hand is
    # moving too fast, and mark them as outliers
    # we use the norm of the acceleration to identify outliers
    hand_acc_norm = np.linalg.norm(cursor.acc, axis=1)
    mask_acceleration = hand_acc_norm > 1500.0
    mask_acceleration = binary_dilation(
        mask_acceleration, structure=np.ones(2, dtype=bool)
    )

    # we also want to identify out of bound segments
    mask_position = np.logical_or(cursor.pos[:, 0] < -10, cursor.pos[:, 0] > 10)
    mask_position = np.logical_or(mask_position, cursor.pos[:, 1] < -10)
    mask_position = np.logical_or(mask_position, cursor.pos[:, 1] > 10)
    # dilate than erode
    mask_position = binary_dilation(mask_position, np.ones(400, dtype=bool))
    mask_position = binary_erosion(mask_position, np.ones(100, dtype=bool))

    outlier_mask = np.logical_or(mask_acceleration, mask_position)

    # convert to interval, you need to find the start and end of the outlier segments
    start = cursor.timestamps[np.where(np.diff(outlier_mask.astype(int)) == 1)[0]]
    if outlier_mask[0]:
        start = np.insert(start, 0, cursor.timestamps[0])

    end = cursor.timestamps[np.where(np.diff(outlier_mask.astype(int)) == -1)[0]]
    if outlier_mask[-1]:
        end = np.append(end, cursor.timestamps[-1])

    cursor_outlier_segments = Interval(start=start, end=end)
    assert cursor_outlier_segments.is_disjoint()
    return cursor_outlier_segments


def split_trials(trials, test_size=0.2, valid_size=0.1, random_state=42):
    num_trials = len(trials)
    train_size = 1.0 - test_size - valid_size

    train_valid_ids, test_ids = train_test_split(
        np.arange(num_trials), test_size=test_size, random_state=random_state
    )
    train_ids, valid_ids = train_test_split(
        train_valid_ids,
        test_size=valid_size / (train_size + valid_size),
        random_state=random_state,
    )

    train_trials = trials.select_by_mask(np.isin(np.arange(num_trials), train_ids))
    valid_trials = trials.select_by_mask(np.isin(np.arange(num_trials), valid_ids))
    test_trials = trials.select_by_mask(np.isin(np.arange(num_trials), test_ids))

    return train_trials, valid_trials, test_trials


def main():
    # use argparse to get arguments from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_dir", type=str, default="./processed")

    args = parser.parse_args()

    brainset_description = BrainsetDescription(
        id="perich_miller_population_2018",
        origin_version="dandi/000688/draft",
        derived_version="1.0.0",
        source="https://dandiarchive.org/dandiset/000688",
        description="This dataset contains electrophysiology and behavioral data from "
        "three macaques performing either a center-out task or a continuous random "
        "target acquisition task. Neural activity was recorded from "
        "chronically-implanted electrode arrays in the primary motor cortex (M1) or "
        "dorsal premotor cortex (PMd) of four rhesus macaque monkeys. A subset of "
        "sessions includes recordings from both regions simultaneously. The data "
        "contains spiking activity—manually spike sorted in three subjects, and "
        "threshold crossings in the fourth subject—obtained from up to 192 electrodes "
        "per session, cursor position and velocity, and other task related metadata.",
    )

    logging.info(f"Processing file: {args.input_file}")

    # open file
    io = NWBHDF5IO(args.input_file, "r")
    nwbfile = io.read()

    # extract subject metadata
    # this dataset is from dandi, which has structured subject metadata, so we
    # can use the helper function extract_subject_from_nwb
    subject = extract_subject_from_nwb(nwbfile)

    # extract experiment metadata
    recording_date = nwbfile.session_start_time.strftime("%Y%m%d")
    device_id = f"{subject.id}_{recording_date}"
    task = (
        "center_out_reaching" if "CO" in args.input_file else "random_target_reaching"
    )
    session_id = f"{device_id}_{task}"

    # register session
    session_description = SessionDescription(
        id=session_id,
        recording_date=datetime.datetime.strptime(recording_date, "%Y%m%d"),
        task=Task.REACHING,
    )

    # register device
    device_description = DeviceDescription(
        id=device_id,
        recording_tech=RecordingTech.UTAH_ARRAY_SPIKES,
    )

    # extract spiking activity
    # this data is from dandi, we can use our helper function
    spikes, units = extract_spikes_from_nwbfile(
        nwbfile, recording_tech=RecordingTech.UTAH_ARRAY_SPIKES
    )

    # extract behavior
    cursor = extract_behavior(nwbfile)
    cursor_outlier_segments = detect_outliers(cursor)

    # extract data about trial structure
    if task == "center_out_reaching":
        trials, movement_phases = extract_center_out_reaching_trials(nwbfile, cursor)
    else:
        trials, movement_phases = extract_random_target_reaching_trials(nwbfile, cursor)

    for key in movement_phases.keys():
        setattr(
            movement_phases,
            key,
            getattr(movement_phases, key).difference(cursor_outlier_segments),
        )

    # close file
    io.close()

    # register session
    data = Data(
        brainset=brainset_description,
        subject=subject,
        session=session_description,
        device=device_description,
        # neural activity
        spikes=spikes,
        units=units,
        # stimuli and behavior
        trials=trials,
        movement_phases=movement_phases,
        cursor=cursor,
        cursor_outlier_segments=cursor_outlier_segments,
        # domain
        domain=cursor.domain,
    )

    # split trials into train, validation and test
    _, valid_trials, test_trials = split_trials(
        trials.select_by_mask(trials.is_valid),
        test_size=0.2,
        valid_size=0.1,
        random_state=42,
    )

    train_sampling_intervals = data.domain.difference(
        (valid_trials | test_trials).dilate(1.0)
    )

    data.set_train_domain(train_sampling_intervals)
    data.set_valid_domain(valid_trials)
    data.set_test_domain(test_trials)

    # save data to disk
    path = os.path.join(args.output_dir, f"{session_id}.h5")
    with h5py.File(path, "w") as file:
        data.to_hdf5(file, serialize_fn_map=serialize_fn_map)


if __name__ == "__main__":
    main()



from brainsets.core import StringIntEnum


class Task(StringIntEnum):
    REACHING = 0

    # For datasets where no tasks are involved
    FREE_BEHAVIOR = 1

    # A Shenoy-style task involving handwriting different characters.
    DISCRETE_WRITING_CHARACTER = 2

    DISCRETE_WRITING_LINE = 3

    CONTINUOUS_WRITING = 4

    # Allen data
    DISCRETE_VISUAL_CODING = 5

    # speech
    DISCRETE_SPEAKING_CVSYLLABLE = 6

    # Full sentence speaking
    CONTINUOUS_SPEAKING_SENTENCE = 7


class Stimulus(StringIntEnum):
    """Stimuli can variously act like inputs (for conditioning) or like outputs."""

    TARGET2D = 0
    TARGETON = 1
    GO_CUE = 2
    TARGETACQ = 3

    DRIFTING_GRATINGS = 4



from enum import IntEnum, auto
import re

import numpy as np


class CVSyllable(IntEnum):
    empty = 0
    baa = auto()
    bee = auto()
    boo = auto()
    chaa = auto()
    chee = auto()
    choo = auto()
    daa = auto()
    dee = auto()
    dhaa = auto()
    dhee = auto()
    dhoo = auto()
    doo = auto()
    faa = auto()
    fee = auto()
    foo = auto()
    gaa = auto()
    gee = auto()
    goo = auto()
    ha = auto()
    haa = auto()
    hee = auto()
    hoo = auto()
    jee = auto()
    kaa = auto()
    kay = auto()
    kee = auto()
    koo = auto()
    laa = auto()
    lee = auto()
    loo = auto()
    maa = auto()
    mee = auto()
    moo = auto()
    naa = auto()
    nee = auto()
    noo = auto()
    paa = auto()
    pee = auto()
    poo = auto()
    raa = auto()
    ree = auto()
    roo = auto()
    saa = auto()
    see = auto()
    shaa = auto()
    she = auto()
    shee = auto()
    shoo = auto()
    skee = auto()
    soo = auto()
    taa = auto()
    tee = auto()
    thaa = auto()
    thee = auto()
    thoo = auto()
    too = auto()
    vaa = auto()
    vee = auto()
    voo = auto()
    waa = auto()
    wee = auto()
    woo = auto()
    ya = auto()
    yaa = auto()
    yee = auto()
    yoo = auto()
    zaa = auto()
    zee = auto()
    zoo = auto()


# From ARPABET
class Phoneme(IntEnum):
    BLANK = 0
    AA = auto()
    AE = auto()
    AH = auto()
    AO = auto()
    AW = auto()
    AY = auto()
    B = auto()
    CH = auto()
    D = auto()
    DH = auto()
    EH = auto()
    ER = auto()
    EY = auto()
    F = auto()
    G = auto()
    HH = auto()
    IH = auto()
    IY = auto()
    JH = auto()
    K = auto()
    L = auto()
    M = auto()
    N = auto()
    NG = auto()
    OW = auto()
    OY = auto()
    P = auto()
    R = auto()
    S = auto()
    SH = auto()
    T = auto()
    TH = auto()
    UH = auto()
    UW = auto()
    V = auto()
    W = auto()
    Y = auto()
    Z = auto()
    ZH = auto()
    SIL = auto()


g2p = None

vocab = "abcdefghijklmnopqrstuvwxyz'- "
# We write phonemes using this pseudo-cypher to calculate phoneme distances
pseudo_code = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMN "


def to_phonemes(sentence, add_interword_symbol=True):
    global g2p
    sentence = "".join(
        [x for x in sentence.lower().strip().replace("--", "") if x in vocab]
    )
    if g2p is None:
        from g2p_en import G2p

        # Grapheme to phoneme library
        g2p = G2p()

    sentence = g2p(sentence)
    phonemes = []
    for p in sentence:
        if add_interword_symbol and p == " ":
            phonemes.append("SIL")
        p = re.sub(r"[0-9]", "", p)  # Remove stress
        if re.match(r"[A-Z]+", p):  # Only keep phonemes
            phonemes.append(p)

    # add one SIL symbol at the end so there's one at the end of each word
    if add_interword_symbol:
        phonemes.append("SIL")

    sentence = np.array([int(Phoneme[x]) for x in phonemes])
    pseudo_sentence = "".join([pseudo_code[x] for x in sentence])
    return pseudo_sentence, sentence


def from_phonemes(sentence):
    # Use the greedy decoding approach, without a language model.
    last_c = -1
    pseudo_transcription = []
    transcription = []
    for c in sentence:
        if c == last_c:
            continue
        last_c = c
        if c == 0:
            continue
        pseudo_transcription.append(pseudo_code[c])
        transcription.append(Phoneme(c).name)
    return ("".join(pseudo_transcription)), (
        ("·".join(transcription)).replace("·SIL·", " ").replace("·SIL", "")
    )



from dataclasses import dataclass

from ..core import StringIntEnum, Dictable


class RecordingTech(StringIntEnum):
    UTAH_ARRAY_SPIKES = 0
    UTAH_ARRAY_THRESHOLD_CROSSINGS = 1
    UTAH_ARRAY_WAVEFORMS = 2
    UTAH_ARRAY_LFPS = 3
    UTAH_ARRAY_AVERAGE_WAVEFORMS = 4
    # As a subordinate category
    UTAH_ARRAY = 9

    NEUROPIXELS_SPIKES = 10
    NEUROPIXELS_THRESHOLDCROSSINGS = 11
    NEUROPIXELS_WAVEFORMS = 12
    NEUROPIXELS_LFPS = 13
    # As a subordinate category
    NEUROPIXELS_ARRAY = 19

    OPENSCOPE_CALCIUM_TRACES = 20
    OPENSCOPE_CALCIUM_RAW = 21

    ECOG_ARRAY_ECOGS = 29
    MICRO_ECOG_ARRAY_ECOGS = 30

    TWO_PHOTON_IMAGING = 40


class Hemisphere(StringIntEnum):
    UNKNOWN = 0
    LEFT = 1
    RIGHT = 2


# @dataclass
# class Channel(Dictable):
#     """Channels are the physical channels used to record the data. Channels are grouped
#     into probes."""

#     id: str
#     local_index: int

#     # Position relative to the reference location of the probe, in microns.
#     relative_x_um: float
#     relative_y_um: float
#     relative_z_um: float

#     area: StringIntEnum
#     hemisphere: Hemisphere = Hemisphere.UNKNOWN


# @dataclass
# class Probe(Dictable):
#     """Probes are the physical probes used to record the data."""

#     id: str
#     type: RecordingTech
#     lfp_sampling_rate: float
#     wideband_sampling_rate: float
#     waveform_sampling_rate: float
#     waveform_samples: int
#     channels: list[Channel]
#     ecog_sampling_rate: float = 0.0



from brainsets.core import StringIntEnum


# from a standard atlas https://github.com/neurodata/neuroparc/blob/master/atlases/label/Human/Anatomical-labels-csv/Glasser.csv
# with a few custmized region
class HomoSapiens(StringIntEnum):
    primary_visual_cortex = 0
    medial_superior_temporal_area = 1
    sixth_visual_area = 2
    second_visual_area = 3
    third_visual_area = 4
    forth_visual_area = 5
    eighth_visual_area = 6
    primary_motor_cortex = 7
    primary_sensory_cortex = 8
    frontal_eye_field = 9
    premotor_eye_field = 10
    area_55b = 11
    area_v3a = 12
    retrosplenial_complex = 13
    parietooccipital_sulcus_area_2 = 14
    seventh_visual_area = 15
    intraparietal_sulcus_area_1 = 16
    fusiform_face_complex = 17
    area_v3b = 18
    lateral_occipital_area_1 = 19
    lateral_occipital_area_2 = 20
    posterior_inferotemporal_complex = 21
    middle_temporal_area = 22
    primary_auditory_cortex = 23
    perisylvian_language_area = 24
    superior_frontal_language_area = 25
    precuneus_visual_area = 26
    superior_temporal_visual_area = 27
    medial_area_7p = 28
    area_7m = 29
    parietooccipital_sulcus_area_1 = 30
    area_23d = 31
    area_ventral_23_a_b = 32
    area_dorsal_23_a_b = 33
    area_31p_ventral = 34
    area_5m = 35
    area_5m_ventral = 36
    area_23c = 37
    area_5l = 38
    dorsal_area_24d = 39
    lateral_area_7a = 40
    supplementary_and_cingulate_eye_field = 41
    area_6m_anterior = 42
    medial_6m_anterior = 43
    medial_area_7a = 44
    lateral_area_7p = 45
    area_7pc = 46
    area_lateral_intraparietal_ventral = 47
    ventral_intraparietal_complex = 48
    medial_intraparietal_area = 49
    area_1 = 50
    area_2 = 51
    area_3a = 52
    dorsal_area_6 = 53
    area_6mp = 54
    ventral_area_6 = 55
    area_posterior_24 = 56
    area_33_prime = 57
    anterior_24_prime = 58
    area_p32_prime = 59
    area_a24 = 60
    area_dorsal_32 = 61
    area_8bm = 62
    area_p32 = 63
    area_10r = 64
    area_47m = 65
    area_8av = 66
    area_8ad = 67
    area_9_middle = 68
    area_8b_lateral = 69
    area_9_posterior = 70
    area_10d = 71
    area_8c = 72
    area_44 = 73
    area_45 = 74
    area_47l = 75
    area_anterior_47r = 76
    rostral_area_6 = 77
    area_ifja = 78
    area_ifjp = 79
    area_ifsp = 80
    area_ifsa = 81
    area_posterior_9_46v = 82
    area_46 = 83
    area_anterior_9_46v = 84
    area_9_46d = 85
    area_9_anterior = 86
    area_10v = 87
    area_anterior_10p = 88
    polar_10p = 89
    area_11 = 90
    area_13 = 91
    orbital_frontal_complex = 92
    area_47s = 93
    area_lateral_intraparietal_dorsal = 94
    area_6_anterior = 95
    inferior_6_8_transitional_area = 96
    superior_6_8_transitional_area = 97
    area_43 = 98
    area_op4_pv = 99
    area_op1_sii = 100
    area_op2_3_vs = 101
    area_52 = 102
    retroinsular_cortex = 103
    area_pfcm = 104
    posterior_insular_area_2 = 105
    area_ta2 = 106
    frontal_opercular_area_4 = 107
    middle_insular_area = 108
    pirform_cortex = 109
    anterior_ventral_insular_area = 110
    anterior_agranular_insula_complex = 111
    frontal_opercular_area_1 = 112
    frontal_opercular_area_3 = 113
    frontal_opercular_area_2 = 114
    area_pft = 115
    anterior_intraparietal_area = 116
    entorhinal_cortex = 117
    presubiculum = 118
    hippocampus = 119
    prostraite_area = 120
    perirhinal_ectorhinal_cortex = 121
    area_stga = 122
    parabelt_complex = 123
    auditory_5_complex = 124
    parahippocampal_area_1 = 125
    parahippocampal_area_3 = 126
    area_stsd_anterior = 127
    area_stsd_posterior = 128
    area_stsv_posterior = 129
    area_tg_dorsal = 130
    area_te1_anterior = 131
    area_te1_posterior = 132
    area_te2_anterior = 133
    area_tf = 134
    area_te2_posterior = 135
    area_pht = 136
    area_ph = 137
    area_temporoparietooccipital_junction_1 = 138
    area_temporoparietooccipital_junction_2 = 139
    area_temporoparietooccipital_junction_3 = 140
    dorsal_transitional_visual_area = 141
    area_pgp = 142
    area_intraparietal_2 = 143
    area_intraparietal_1 = 144
    area_intraparietal_0 = 145
    area_pf_opercular = 146
    area_pf_complex = 147
    area_pfm_complex = 148
    area_pgi = 149
    area_pgs = 150
    area_v6a = 151
    ventromedial_visual_area_1 = 152
    ventromedial_visual_area_3 = 153
    parahippocampal_area_2 = 154
    area_v4t = 155
    area_fst = 156
    area_v3cd = 157
    area_lateral_occipital_3 = 158
    ventromedial_visual_area_2 = 159
    area_31pd = 160
    area_31a = 161
    ventral_visual_complex = 162
    area_25 = 163
    area_s32 = 164
    posterior_ofc_complex = 165
    area_posterior_insular_1 = 166
    insular_granular_complex = 167
    area_frontal_opercular_5 = 168
    area_posterior_10p = 169
    area_posterior_47r = 170
    area_tg_ventral = 171
    medial_belt_complex = 172
    lateral_belt_complex = 173
    auditory_4_complex = 174
    area_stsv_anterior = 175
    area_te1_middle = 176
    parainsular_area = 177
    area_anterior_32_prime = 178

    # customized
    ventral_sensorimotor_cortex = 179



import argparse
import datetime
import logging
import os
from pathlib import Path

import h5py
import numpy as np
import pandas as pd

from brainsets import serialize_fn_map
from brainsets.descriptions import (
    BrainsetDescription,
    SessionDescription,
    SubjectDescription,
    DeviceDescription,
)

from brainsets.taxonomy import (
    RecordingTech,
    Species,
    Task,
)
from temporaldata import (
    ArrayDict,
    Data,
    Interval,
    IrregularTimeSeries,
)

logging.basicConfig(level=logging.INFO)


def extract_behavior(h5file):
    """Extract the behavior from the h5 file.

    ..note::
        Cursor position and target position are in the same frame of reference.
        They are both of size (sequence_len, 2). Finger position can be either 3d or 6d,
        depending on the sequence.
    """

    cursor_pos = h5file["cursor_pos"][:].T
    finger_pos = h5file["finger_pos"][:].T
    target_pos = h5file["target_pos"][:].T
    timestamps = h5file["t"][:][0]

    expected_period = 0.004
    assert np.all(np.abs(np.diff(timestamps) - expected_period) < 1e-4)

    # calculate the velocity of the cursor
    cursor_vel = np.gradient(cursor_pos, timestamps, edge_order=1, axis=0)
    cursor_acc = np.gradient(cursor_vel, timestamps, edge_order=1, axis=0)
    finger_vel = np.gradient(finger_pos, timestamps, edge_order=1, axis=0)

    cursor = IrregularTimeSeries(
        timestamps=timestamps,
        pos=cursor_pos,
        vel=cursor_vel,
        acc=cursor_acc,
        domain="auto",
    )

    # The position of the working fingertip in Cartesian coordinates (z, -x, -y), as
    # reported by the hand tracker in cm. Thus the cursor position is an affine
    # transformation of fingertip position.
    finger = IrregularTimeSeries(
        timestamps=timestamps,
        pos_3d=finger_pos[:, :3],
        vel_3d=finger_vel[:, :3],
        domain="auto",
    )
    if finger_pos.shape[1] == 6:
        finger.orientation = finger_pos[:, 3:]
        finger.angular_vel = finger_vel[:, 3:]

    assert cursor.is_sorted()
    assert finger.is_sorted()

    return cursor, finger


def detect_no_movement(cursor):
    """Detect segments where the cursor is not moving"""
    mask = np.linalg.norm(cursor.vel, axis=1) < 10.0

    # convert to interval, you need to find the start and end of the outlier segments
    start = cursor.timestamps[np.where(np.diff(mask.astype(int)) == 1)[0]]
    if mask[0]:
        start = np.insert(start, 0, cursor.timestamps[0])

    end = cursor.timestamps[np.where(np.diff(mask.astype(int)) == -1)[0]]
    if mask[-1]:
        end = np.append(end, cursor.timestamps[-1])

    no_movement_segments = Interval(start=start, end=end)
    assert no_movement_segments.is_disjoint()

    no_movement_segments = no_movement_segments.select_by_mask(
        (no_movement_segments.end - no_movement_segments.start) > 10.0
    )
    return no_movement_segments


def extract_spikes(h5file: h5py.File):
    r"""This dataset has a mixture of sorted and unsorted (threshold crossings)
    units.
    """

    # helpers specific to spike extraction
    def _to_ascii(vector):
        return ["".join(chr(char) for char in row) for row in vector]

    def _load_references_2d(h5file, ref_name):
        return [[h5file[ref] for ref in ref_row] for ref_row in h5file[ref_name][:]]

    spikesvec = _load_references_2d(h5file, "spikes")
    waveformsvec = _load_references_2d(h5file, "wf")

    # this is slightly silly but we can convert channel names back to an ascii token
    # this way.
    chan_names = _to_ascii(
        np.array(_load_references_2d(h5file, "chan_names")).squeeze()
    )

    spikes = []
    unit_index = []
    types = []
    waveforms = []
    unit_meta = []

    # The 0'th spikesvec corresponds to unsorted thresholded units, the rest are sorted.
    suffixes = ["unsorted"] + [f"sorted_{i:02}" for i in range(1, 11)]
    type_map = [int(RecordingTech.UTAH_ARRAY_THRESHOLD_CROSSINGS)] + [
        int(RecordingTech.UTAH_ARRAY_SPIKES)
    ] * 10

    encountered = set()

    unit_index_delta = 0
    for j in range(len(spikesvec)):
        crossings = spikesvec[j]
        for i in range(len(crossings)):
            spiketimes = crossings[i][:][0]
            if spiketimes.ndim == 0:
                continue

            spikes.append(spiketimes)
            area, channel_number = chan_names[i].split(" ")

            unit_name = f"{chan_names[i]}/{suffixes[j]}"

            unit_index.append([unit_index_delta] * len(spiketimes))
            types.append(np.ones_like(spiketimes, dtype=np.int64) * type_map[j])

            if unit_name in encountered:
                raise ValueError(f"Duplicate unit name: {unit_name}")
            encountered.add(unit_name)

            wf = np.array(waveformsvec[j][i][:])
            unit_meta.append(
                {
                    "count": len(spiketimes),
                    "channel_name": chan_names[i],
                    "id": unit_name,
                    "area_name": area,
                    "channel_number": channel_number,
                    "unit_number": j,
                    "type": type_map[j],
                    "average_waveform": wf.mean(axis=1)[:48],
                }
            )
            waveforms.append(wf.T)
            unit_index_delta += 1

    spikes = np.concatenate(spikes)
    waveforms = np.concatenate(waveforms)
    unit_index = np.concatenate(unit_index)

    spikes = IrregularTimeSeries(
        timestamps=spikes,
        unit_index=unit_index,
        waveforms=waveforms,
        domain="auto",
    )
    spikes.sort()

    units = ArrayDict.from_dataframe(pd.DataFrame(unit_meta))
    return spikes, units


def split_intervals(data):
    # slice the session into 10 blocks then randomly split them into train,
    # validation and test sets, using a 8/1/1 ratio.
    task_domain = data.domain
    if len(data.no_movement_segments) > 0:
        task_domain = task_domain.difference(data.no_movement_segments)
        task_domain = task_domain.select_by_mask(
            task_domain.end - task_domain.start > 10.0
        )

    intervals = Interval.linspace(task_domain.start[0], task_domain.end[-1], 10)
    if len(data.no_movement_segments) > 0:
        task_ratio = []
        for start, end in zip(intervals.start, intervals.end):
            intersection = task_domain & Interval(start, end)
            duration = np.sum(intersection.end - intersection.start)
            task_ratio.append(duration / (end - start))

        task_ratio = np.array(task_ratio)
        task_ratio_with_index = np.column_stack((np.arange(len(intervals)), task_ratio))
        rng = np.random.default_rng(42)
        rng.shuffle(task_ratio_with_index)
        sorted_index = task_ratio_with_index[task_ratio_with_index[:, 1].argsort()][
            :, 0
        ].astype(int)

        train_sampling_intervals = Interval(
            start=intervals.start[sorted_index[:8]], end=intervals.end[sorted_index[:8]]
        )
        valid_sampling_intervals = Interval(
            start=intervals.start[sorted_index[8:9]],
            end=intervals.end[sorted_index[8:9]],
        )
        test_sampling_intervals = Interval(
            start=intervals.start[sorted_index[9:]], end=intervals.end[sorted_index[9:]]
        )

        train_sampling_intervals.sort()
        valid_sampling_intervals.sort()
        test_sampling_intervals.sort()

        train_sampling_intervals = train_sampling_intervals.difference(
            data.no_movement_segments
        )
        valid_sampling_intervals = valid_sampling_intervals.difference(
            data.no_movement_segments
        )
        test_sampling_intervals = test_sampling_intervals.difference(
            data.no_movement_segments
        )

        train_sampling_intervals = train_sampling_intervals.select_by_mask(
            train_sampling_intervals.end - train_sampling_intervals.start > 10.0
        )
        valid_sampling_intervals = valid_sampling_intervals.select_by_mask(
            valid_sampling_intervals.end - valid_sampling_intervals.start > 10.0
        )
        test_sampling_intervals = test_sampling_intervals.select_by_mask(
            test_sampling_intervals.end - test_sampling_intervals.start > 10.0
        )
    else:
        [
            train_sampling_intervals,
            valid_sampling_intervals,
            test_sampling_intervals,
        ] = intervals.split([8, 1, 1], shuffle=True, random_seed=42)
    return train_sampling_intervals, valid_sampling_intervals, test_sampling_intervals


def main():
    # use argparse to get arguments from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_file", type=str)
    parser.add_argument("--output_dir", type=str, default="./processed")

    args = parser.parse_args()

    brainset_description = BrainsetDescription(
        id="odoherty_sabes_nonhuman_2017",
        origin_version="583331",  # Zenodo version
        derived_version="1.0.0",
        source="https://zenodo.org/record/583331",
        description="The behavioral task was to make self-paced reaches to targets "
        "arranged in a grid (e.g. 8x8) without gaps or pre-movement delay intervals. "
        "One monkey reached with the right arm (recordings made in the left hemisphere)"
        "The other reached with the left arm (right hemisphere). In some sessions "
        "recordings were made from both M1 and S1 arrays (192 channels); "
        "in most sessions M1 recordings were made alone (96 channels).",
    )

    logging.info(f"Processing file: {args.input_file}")

    # open file
    h5file = h5py.File(args.input_file, "r")

    # extract experiment metadata
    # determine session_id and sortset_id
    session_id = Path(args.input_file).stem  # type: ignore
    device_id = session_id[:-3]
    assert device_id.count("_") == 1, f"Unexpected file name: {device_id}"

    animal, recording_date = device_id.split("_")
    subject = SubjectDescription(
        id=animal,
        species=Species.MACACA_MULATTA,
    )

    session_description = SessionDescription(
        id=session_id,
        recording_date=datetime.datetime.strptime(recording_date, "%Y%m%d"),
        task=Task.REACHING,
    )

    device_description = DeviceDescription(
        id=device_id,
        recording_tech=RecordingTech.UTAH_ARRAY_SPIKES,
    )

    # extract spiking activity, unit metadata and channel names info
    spikes, units = extract_spikes(h5file)

    # extract behavior
    cursor, finger = extract_behavior(h5file)
    no_movement_segments = detect_no_movement(cursor)

    # close file
    h5file.close()

    # register session
    data = Data(
        brainset=brainset_description,
        subject=subject,
        session=session_description,
        device=device_description,
        # neural activity
        spikes=spikes,
        units=units,
        # stimuli and behavior
        cursor=cursor,
        finger=finger,
        no_movement_segments=no_movement_segments,
        domain=cursor.domain,
    )

    train_sampling_intervals, valid_sampling_intervals, test_sampling_intervals = (
        split_intervals(data)
    )
    # set the domains
    data.set_train_domain(train_sampling_intervals)
    data.set_valid_domain(valid_sampling_intervals)
    data.set_test_domain(test_sampling_intervals)

    # save data to disk
    path = os.path.join(args.output_dir, f"{session_id}.h5")
    with h5py.File(path, "w") as file:
        data.to_hdf5(file, serialize_fn_map=serialize_fn_map)


if __name__ == "__main__":
    main()
